/*
  This code reads IMU data, which is received via a nRF24L01+PA+LNA module, publishes the data to ROS to calculate the velocity which the robot should drive. The command velocity will be read by the Arduino, by subscribing to the ROS topic. As an additional feature (when the publishWheelSpeedEstimate variable is set to true) the code reads the speed of the motor wheels, computes the linear velocity and rotation speed of the robot and published the data to ROS.
*/

//include the necessary libraries
#include <SPI.h>
#include <nRF24L01.h>
#include <RF24.h>
#include <HardwareSerial.h>
#include <SoftwareSerial.h>
#include "ODriveArduino.h"
#include <ros.h>                
#include <std_msgs/String.h>
//#include <ros/time.h>                
#include <sensor_msgs/Imu.h>           
#include <sensor_msgs/MagneticField.h> 
#include <geometry_msgs/Twist.h>  


//ODrive Controller
  HardwareSerial& odrive_serial = Serial2;
  ODriveArduino odrive(odrive_serial);
  double timeTransmissionCounter = 0;

  double velEstimateRightWheel = 0;
  double velEstimateLeftWheel = 0;
  double estimatedLinVel = 0;
  double estimatedRotSpeed = 0;

//variables needed for the nRF24L01+ module 
  RF24 radio(7, 8); // CE, CSN
  const byte address[6] = "00001";
  
  struct Data_Package {  
    float ax,ay,az;
    float gx,gy,gz;
  };
  
  Data_Package data; 

//Connection between the Arduino and ROS
  ros::NodeHandle  nh;  
  //for publishing the accelerometer and angular velocity data
  sensor_msgs::Imu imu_msg;
  ros::Publisher sensorDataPublisher( "imu/data_raw", &imu_msg);
 
  geometry_msgs::Twist stopCmdVelMsg;
  ros::Publisher cmdVelPublisher( "cmd_vel", &stopCmdVelMsg);
  float pi = 3.1415926535897932384626433832795;

  geometry_msgs::Twist measured_vel_msg;
  ros::Publisher measuredVelPublisher( "measured_vel", &measured_vel_msg);
  boolean publishWheelSpeedEstimate = true;   

boolean imuTeleoperationMode = false; 
//the counter is used to publish specific messages for a n-times to ROS
int counter = 0;

//callbackfunction (communication between ROS and Arduino)
  char rightWheelSpeed[25];
  char leftWheelSpeed[25];
  char linearVelocity[25];
  char angularRotSpeed[25];
  double linearVelocityDouble;
  double linearVelocityRightWheel;
  double linearVelocityLeftWheel;
  double angularRotSpeedDouble;
  int maxRotationSpeed = 3;   

//the callback-function 
void messageCb( const geometry_msgs::Twist& cmd_vel_msg){

   //save the cmd_vel topic in variables
  linearVelocityDouble = cmd_vel_msg.linear.x;  
  angularRotSpeedDouble = cmd_vel_msg.angular.z; 

 //checking that the data from the cmd_vel message is not NAN 
  if(!isnan(linearVelocityDouble) && !isnan(angularRotSpeedDouble)){ 
   
    linearVelocityRightWheel = 1.928 * linearVelocityDouble;  
    linearVelocityLeftWheel = 1.928 * linearVelocityDouble;
  
    //superpositioning the rotation speed with linear speed
    linearVelocityRightWheel = linearVelocityRightWheel + 0.15 * angularRotSpeedDouble;
    linearVelocityLeftWheel = linearVelocityLeftWheel - 0.15 * angularRotSpeedDouble;
  
  
    //limiting the motor speed to the allowed area
    if(linearVelocityRightWheel >= maxRotationSpeed){
      linearVelocityRightWheel = maxRotationSpeed;
    }
    if(linearVelocityLeftWheel >= maxRotationSpeed){
      linearVelocityLeftWheel = maxRotationSpeed;
    }
    if(linearVelocityRightWheel <= -maxRotationSpeed){
      linearVelocityRightWheel = -maxRotationSpeed;
    }
    if(linearVelocityLeftWheel <= -maxRotationSpeed){
      linearVelocityLeftWheel = -maxRotationSpeed;
    }
    
    //the linear velocity of the right and left wheel are converted from the type double to the type char-Array, in order to print it on the terminal
    dtostrf(linearVelocityRightWheel,20,19,rightWheelSpeed);
    dtostrf(linearVelocityLeftWheel,20,19,leftWheelSpeed);  
    
    //printing the rotation speed of the wheels in the terminal
    if(timeTransmissionCounter < 100){ 
      nh.loginfo("---------------------------------------");
      nh.loginfo("right wheel");
      nh.loginfo(rightWheelSpeed);
      nh.loginfo("left wheel");
      nh.loginfo(leftWheelSpeed);
      nh.loginfo("---------------------------------------");
    }
    
   //if the nRF24L01+ modules are not connected with each other for a short periof of time, the motor speed should be set to 0 
    if(timeTransmissionCounter < 100){      
      //setting the velocity of the wheels:
      odrive.SetVelocity(0, -linearVelocityLeftWheel);
      odrive.SetVelocity(1, linearVelocityRightWheel);
    } else {
      //stopping the motors due to loss of connection
      odrive.SetVelocity(0, 0);    
      odrive.SetVelocity(1, 0);
    }
  } else {  //the cmd_vel message is NAN -> the motor speed will be set to 0
      odrive.SetVelocity(0, 0);    
      odrive.SetVelocity(1, 0);
      nh.loginfo("the cmd_vel Message includes NAN-Values - probably there is an issue with the complementaryFilter");
  } 

  nh.spinOnce();
}

  //defining the Arduino as a Subscriber of the cmd_vel topic
  ros::Subscriber<geometry_msgs::Twist> sub("cmd_vel", &messageCb ); 

void setup() {
  nh.getHardware()-> setBaud(115200);

  //ROS-Communication Setup
     nh.initNode();                           
     nh.advertise(sensorDataPublisher);       
     nh.advertise(measuredVelPublisher);   
     nh.advertise(cmdVelPublisher);        
     nh.subscribe(sub);          

  // Serial to PC
  //Serial.begin(115200);

  //setup for the nRF24L01 module
    radio.begin();
    radio.openReadingPipe(0, address);
    radio.setPALevel(RF24_PA_MIN); 
    radio.startListening();
 
  // ODrive baudrate
  odrive_serial.begin(115200);
}


void loop() {

  if(!nh.connected()){  
     odrive.SetVelocity(0, 0);     
     odrive.SetVelocity(1, 0);   
     nh.loginfo("motorspeed is set to 0 due to the missing connection to the PC.");  
  } else {

    if (radio.available()) {
      imuTeleoperationMode = true;
      timeTransmissionCounter = 0;  
      radio.read(&data, sizeof(data));
   
    //sending the raw IMU data to ROS 
      imu_msg.header.frame_id = "IMU msg";
      imu_msg.header.stamp = nh.now();
    
      //setting the remaining message data  
      //as the orientation is not known, the covariance is -1 
      imu_msg.orientation_covariance[0] = -1;
      imu_msg.orientation_covariance[1] = -1;
      imu_msg.orientation_covariance[2] = -1;
      imu_msg.orientation_covariance[3] = -1;
      imu_msg.orientation_covariance[4] = -1;
      imu_msg.orientation_covariance[5] = -1;
      imu_msg.orientation_covariance[6] = -1;
      imu_msg.orientation_covariance[7] = -1;
      imu_msg.orientation_covariance[8] = -1;
  
     //converting the unis grad/s to rad/s
      float gxRad = data.gx * (2*pi)/360;
      float gyRad = data.gy * (2*pi)/360;
      float gzRad = data.gz * (2*pi)/360;
  
      //setting the rotationspeed within the message
      imu_msg.angular_velocity.x = gxRad;
      imu_msg.angular_velocity.y = gyRad;
      imu_msg.angular_velocity.z = gzRad;
    
      //setting the acceleration within the message (in m/s^2) 
      imu_msg.linear_acceleration.x = data.ax;
      imu_msg.linear_acceleration.y = data.ay;
      imu_msg.linear_acceleration.z = data.az;
  
      //publishing the message to ROS
      sensorDataPublisher.publish( &imu_msg );
    
    } else if(radio.available() == false && imuTeleoperationMode == true) { 
       
      //if there is no radio connection for a short period of time the motors should be stopped
      delay(1);
    
      if(timeTransmissionCounter >= 100){
        nh.loginfo("The motor stop sequence is started... ---------------");
        odrive.SetVelocity(0, 0);
        odrive.SetVelocity(1, 0);
        imuTeleoperationMode = false;     
        nh.loginfo("no radio connection ...");
  
        //setting up the header
        imu_msg.header.frame_id = "MyArduinoImuNode";
        imu_msg.header.stamp = nh.now();
      
        //setting the remaining message data  
        //as the orientation is not known, the covariance is -1 
        imu_msg.orientation_covariance[0] = -1;
        imu_msg.orientation_covariance[1] = -1;
        imu_msg.orientation_covariance[2] = -1;
        imu_msg.orientation_covariance[3] = -1;
        imu_msg.orientation_covariance[4] = -1;
        imu_msg.orientation_covariance[5] = -1;
        imu_msg.orientation_covariance[6] = -1;
        imu_msg.orientation_covariance[7] = -1;
        imu_msg.orientation_covariance[8] = -1;
        
        //setting the rotation speed within the message
        imu_msg.angular_velocity.x = 0.0;
        imu_msg.angular_velocity.y = 0.0;
        imu_msg.angular_velocity.z = 0.0;
      
        //setting the acceleration within the message (in m/s^2) 
        imu_msg.linear_acceleration.x = 0.0;
        imu_msg.linear_acceleration.y = 0.0;
        imu_msg.linear_acceleration.z = 9.81;
  
        //publishing the message
        sensorDataPublisher.publish( &imu_msg );

        counter = 0;
        while (counter < 20){       
          //the motorspeed is set to 0 in the loop, so that the motors directly stop instead of slowing down
          odrive.SetVelocity(0, 0);    
          odrive.SetVelocity(1, 0);    
          sensorDataPublisher.publish( &imu_msg );
          counter ++;
          nh.spinOnce();
        }

       counter = 0;
        while (counter < 50){
          cmdVelPublisher.publish(&stopCmdVelMsg);
          nh.spinOnce();
          counter ++;
        }
        nh.loginfo("The motor stop sequence is finished. The motors can now be controlled via other devices which publish to the cmd_vel topic.");
      }
    }
  }

  //if the publishWheelSpeedEstimate variable is set to true, the Arduino will read the speed of the motors, calculate the robot linear velocity and rotation speed and publish it to ROS
  if(publishWheelSpeedEstimate == true){

      //reading the speed of the individual motor
      velEstimateLeftWheel = -odrive.GetVelocity(0);
      velEstimateRightWheel = odrive.GetVelocity(1);
      
      //converting the units to meters per second
      // 1 turn/s = 0.5187 m/s
      velEstimateLeftWheel = 0.5187 * velEstimateLeftWheel; 
      velEstimateRightWheel = 0.5187 * velEstimateRightWheel; 
    
      //computing the linear velocity & the rotation speed of the robot
      estimatedLinVel = velEstimateLeftWheel/2 + velEstimateRightWheel/2;
      estimatedRotSpeed = (velEstimateRightWheel - velEstimateLeftWheel)/0.3; 

      //setting up the message, which publishes the linear velocity and the rotation speed to the measured_vel topic
      measured_vel_msg.linear.x = estimatedLinVel;
      measured_vel_msg.linear.y = 0;
      measured_vel_msg.linear.z = 0;

      //setting the rotation speed within the message
      measured_vel_msg.angular.x = 0;
      measured_vel_msg.angular.y = 0;
      measured_vel_msg.angular.z = estimatedRotSpeed;
    
      //publishing the message to ROS
      measuredVelPublisher.publish( &measured_vel_msg );
  }

  nh.spinOnce();

}

