/*
This sketch uses RC transmitter and receiver to control the robot. The configuration is based on LuSeka HoverBot implementation at https://github.com/LuSeKa/HoverBot/blob/master/HoverBot.ino
*/

// Include necessary libraries
#include <ODriveArduino.h>
#include <Metro.h>

// RC receiver channel connection to Arduino
#define CH1 2
#define CH2 3
#define CH3 18 

// LED indicator
#define LEDPIN 13

//Define the motors direction
#define M0_direction -1
#define M1_direction 1

// Define the RC controller tunning paramters
#define kp_throttle 0.0011
#define kp_steering 0.001

// Define the radio control activation threshold and centre
#define activation_threshold 1500
#define pwm_centre 1500

// Define the blinking, controller and activation intervals
#define blink_interval 200
#define display_interval 20
#define controller_interval 10
#define activation_interval 50

// Baudrate
#define pc_baudrate 115200
#define odrive_baudrate 115200

// Initialize the pwm decoder and the motor variables
int throttle = 0;
int steering = 0;
int pwm_mode = 0;
int motors_active_state = 0;

 // Intitialize Odrive Arduino Serial Communication
ODriveArduino odrive(Serial2);

// Interrupt callbacks to measure the pwm duty cycles

void decode_CH1() {
  static unsigned long rise_duration = micros();
  if (digitalRead(CH1) == 1) {
    rise_duration = micros();
  }
  else {
    steering = micros() - rise_duration;
  }
}

void decode_CH2() {
  static unsigned long rise_duration = micros();
  if (digitalRead(CH2) == 1) {
    rise_duration = micros();
  }
  else { 
    throttle = micros() - rise_duration;
  }
}

void decode_CH3() {
  static unsigned long rise_duration = micros();
  if (digitalRead(CH3) == 1) {
    rise_duration = micros();
  }
  else {
    pwm_mode = micros() - rise_duration;
  }
}

Metro LED_blink = Metro(blink_interval);
Metro display_out = Metro(display_interval);
Metro controller = Metro(controller_interval);
Metro activation = Metro(activation_interval);

void setup() {
  pinMode(LEDPIN, OUTPUT);
  Serial.begin(pc_baudrate);
  Serial2.begin(odrive_baudrate);
  
  // pwm decoder interrupts
  attachInterrupt(digitalPinToInterrupt(CH1), decode_CH1, CHANGE);
  attachInterrupt(digitalPinToInterrupt(CH2), decode_CH2, CHANGE);
  attachInterrupt(digitalPinToInterrupt(CH3), decode_CH3, CHANGE);
}

// show that the sketch is running, plot from the arduino serial plotter, loop the control and activation tasks
void loop() {
  blink_task();
  display_task();
  control_task();
  activation_task();
}

void blink_task() {
  if (LED_blink.check()) {
    digitalWrite(LEDPIN, !digitalRead(LEDPIN));
  }
}

void display_task() {
  if (display_out.check()) {
    Serial.print(throttle);
    Serial.print('\t');
    Serial.print(steering);
    Serial.print('\t');
    Serial.println(pwm_mode);
  }
}

void control_task() {
  if (controller.check()) {
    movement_control();
  }
}

void activation_task() {
  if (activation.check()) {
    switch_mode(pwm_mode > activation_threshold);
  }
}

void movement_control() {
  
  // Position and steering control
  float position_control = kp_throttle * (throttle - pwm_centre);
  float steering_control = kp_steering * (steering - pwm_centre);  

  float move_right = position_control + steering_control;
  float move_left  = position_control - steering_control;
  
  odrive.SetVelocity(0, M0_direction * move_right);
  odrive.SetVelocity(1, M1_direction * move_left);
}

void switch_mode(int desired_mode) {
   if (desired_mode == motors_active_state) {
  }
  else {
    if (desired_mode == 1) {
     int requested_state = AXIS_STATE_CLOSED_LOOP_CONTROL;
      Serial.println("Activating_motors:_Make_sure_the_robot_is_free_to_move");
      odrive.run_state(0, requested_state, false);
      odrive.run_state(1, requested_state, false);
    }
    else if (desired_mode == 0) {
     int requested_state = AXIS_STATE_IDLE;
      Serial.println("De-activating_motors");
      odrive.run_state(0, requested_state, false);
      odrive.run_state(1, requested_state, false);
    }
    else {
      Serial.println("Invalid mode selection");
    }
    motors_active_state = desired_mode;
  }
  return;
}
